#ifndef TEST_STANDALONE
#include <intercept.h>
#endif

#include <sys/time.h>
#include <sys/select.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <stdarg.h>
#include <assert.h>

#include <gtk/gtk.h>
#include <glib.h>

static size_t indent = 0;
static GHashTable* widget_count;
static GHashTable* internal_children;
static FILE* output_file;

static void register_internal_children(GtkWidget *widget)
{
#define REGISTER(type, name) g_hash_table_replace(internal_children, type(widget)->name, #name);

	if (GTK_IS_COMBO(widget))
	{
		REGISTER(GTK_COMBO, entry);
		REGISTER(GTK_COMBO, button);
		REGISTER(GTK_COMBO, popup);
		REGISTER(GTK_COMBO, popwin);
		REGISTER(GTK_COMBO, list);
	}

	if (GTK_IS_FONT_SELECTION_DIALOG(widget))
	{
		REGISTER(GTK_FONT_SELECTION_DIALOG, ok_button);
		REGISTER(GTK_FONT_SELECTION_DIALOG, cancel_button);
		REGISTER(GTK_FONT_SELECTION_DIALOG, apply_button);
		g_hash_table_replace(internal_children, GTK_FONT_SELECTION_DIALOG(widget)->fontsel, "font_selection");
	}		

	if (GTK_IS_IMAGE_MENU_ITEM(widget))
	{
		GtkWidget* image = gtk_image_menu_item_get_image(GTK_IMAGE_MENU_ITEM(widget));

		if (image)
			g_hash_table_replace(internal_children, image, "image");
	}

	if (GTK_IS_SCROLLED_WINDOW(widget))
	{
		REGISTER(GTK_SCROLLED_WINDOW, vscrollbar);
		REGISTER(GTK_SCROLLED_WINDOW, hscrollbar);
	}

	if (GTK_IS_FILE_SELECTION(widget))
	{
		REGISTER(GTK_FILE_SELECTION, ok_button);
		REGISTER(GTK_FILE_SELECTION, cancel_button);
		REGISTER(GTK_FILE_SELECTION, help_button);
	}

	if (GTK_IS_COLOR_SELECTION_DIALOG(widget))
	{
		REGISTER(GTK_COLOR_SELECTION_DIALOG, ok_button);
		REGISTER(GTK_COLOR_SELECTION_DIALOG, cancel_button);
		REGISTER(GTK_COLOR_SELECTION_DIALOG, help_button);
		g_hash_table_replace(internal_children, GTK_COLOR_SELECTION_DIALOG(widget)->colorsel, "color_selection");
	}

	if (GTK_IS_DIALOG(widget))
	{
		REGISTER(GTK_DIALOG, vbox);
		REGISTER(GTK_DIALOG, action_area);
	}
#undef REGISTER
}

static void print(const char* st, ...)
{
	va_list args;
	size_t i;

	for (i = 0; i < indent; ++i)
		fprintf(output_file, "  ");
	
	va_start(args, st);
	vfprintf(output_file, st, args);
	va_end(args);
	
	fprintf(output_file, "\n");
	/* I flush just in case I segfault, I can get a clue about where the problem lies */
	fflush(output_file);
}

/**
 * Should convert from a string as GtkTextView to text_viewx,
 * where x is an incremental counter (specific to each name)
 */
static char* get_object_name(const char* widget_name)
{
	char* object_name;
	char* tmp;
	guint nb;

	g_return_val_if_fail(widget_name, NULL);
	g_return_val_if_fail(*widget_name, NULL);

	/* remove Gtk prefix */
	if (strncmp(widget_name, "Gtk", 3) == 0)
		widget_name += 3;

	/* remove Gnome prefix */
	if (strncmp(widget_name, "Gnome", 5) == 0)
		widget_name += 5;

	object_name = (char*) g_malloc(strlen(widget_name) * 2);
	if (object_name)
	{
		tmp = object_name;
		
		/* copy the first letter */
		*tmp++ = tolower(*widget_name++);

		/* duplicate the original name, changing the uppercased letters by
		 * '_' + lower(ch) */
		for (; *widget_name; ++widget_name)
		{
			if (isupper(*widget_name))
			{
				*tmp++ = '_';
				*tmp++ = tolower(*widget_name);
			}
			else
				*tmp++ = *widget_name;
		}

		*tmp = 0;

		nb = GPOINTER_TO_UINT(g_hash_table_lookup(widget_count, object_name));
		g_hash_table_replace(widget_count, g_strdup(object_name), GUINT_TO_POINTER(++nb));
		tmp = g_strdup_printf("%s%d", object_name, nb);

		g_free(object_name);
		object_name = tmp;
	}

	return object_name;
}

/* you should g_free this value */
static char* convert_property(const char* value)
{
	char* tmp1;
	char* tmp2;

	if (!value)
		return g_strdup("");
	
	/* change '-' by '_' */
	tmp2 = tmp1 = g_strdup(value);
	do
	{
		if (*tmp2 == '-')
			*tmp2 = '_';
	}
	while (*tmp2++);

	return tmp1;
}

/* you should g_free this value */
/* TODO: I'm 99% confident that this function is broken XML-wise.
 * But by now it's good enough to quote the little tags embedded in the GtkLabels */
static char* xml_quote(const char* value)
{
	size_t size = 256;
	char* retval = NULL;
	const char* it;
	char* it2;
	
	if (!value)
		return g_strdup("");
	
	retval = g_malloc(size);
	if (!retval)
		return NULL;

	/* now we should xml-quote the text (changing '<' by &lt; and so on) */
	it2 = retval;
	for (it = value; *it != 0; it = g_utf8_next_char(it))
	{
		/* if we run out of space in the current string, allocate a new, bigger, one
		 * (TODO: consider using GString)
		 * "-7" due to:
		 *     1) maximal size in bytes of a UTF-8 character (6 bytes)
		 *     2) the '\0' that we should add if we run out of space
		 */
		if ((size_t) (it2 - retval) > size - 7)
		{
			char* tmp = g_malloc(size);
			if (!tmp)
			{
				g_free(it2);
				return NULL;
			}

			*it2 = 0;
			strcpy(tmp, retval);
			g_free(retval);
			
			size *= 2;
			retval = g_malloc(size);
		}

		switch (*it)
		{
		case '<':
			*it2++ = '&';
			*it2++ = 'l';
			*it2++ = 't';
			*it2++ = ';';
			break;

		case '>':
			*it2++ = '&';
			*it2++ = 'g';
			*it2++ = 't';
			*it2++ = ';';
			break;
			
		case '&':
			*it2++ = '&';
			*it2++ = 'a';
			*it2++ = 'm';
			*it2++ = 'p';
			*it2++ = ';';
			break;

		default:
			*it2 = *it;
			it2 = g_utf8_next_char(it2);
		}
	}
	*it2 = 0;

	return retval;
}

/* you should g_free this value */
static char* convert_value(const char* value)
{
	if (!value)
		return g_strdup("");
	
	/* if it starts by GTK_[EF] */
	if (value[0] == 'G')
	{
		gboolean expand = FALSE;
		gboolean fill = FALSE;

		if (strncmp(value, "GTK_FILL", 8) == 0)
		{
			fill = TRUE;
			value += 8;
		}
		else if (strncmp(value, "GTK_EXPAND", 10) == 0)
		{
			expand = TRUE;
			value += 10;
		}

		while (*value == ' ' || *value == '|')
			++value;
		
		if (strncmp(value, "GTK_FILL", 8) == 0)
			fill = TRUE;
		else if (strncmp(value, "GTK_EXPAND", 10) == 0)
			expand = TRUE;

		if (expand || fill)
		{
			if (expand && fill)
				return g_strdup("expand|fill");
			else if (expand)
				return g_strdup("expand");
			else
				return g_strdup("fill");
		}
	}
	
	return xml_quote(value);
}

static void dump_packing_properties(GtkWidget* widget, GParamSpec** spec, guint nproperties)
{
	guint i;
	const char* prop_name;
	const char* value_content;
	char* tmp1;
	char* tmp2;
	GtkWidget* parent = gtk_widget_get_parent(widget);
	
	for (i = 0; i < nproperties; ++i)
	{
		GValue value = { 0, };
		prop_name = g_param_spec_get_name(spec[i]);
		
		g_value_init(&value, spec[i]->value_type);
		gtk_container_child_get_property(GTK_CONTAINER(parent), widget, prop_name, &value);

		if (g_value_type_transformable(G_VALUE_TYPE(&value), G_TYPE_STRING))
		{
			GValue tmp_value = { 0, };

			g_value_init(&tmp_value, G_TYPE_STRING);
			g_value_transform(&value, &tmp_value);
			value_content = g_value_get_string(&tmp_value);

			tmp1 = convert_property(prop_name);
			tmp2 = convert_value(value_content);
			print("<property name=\"%s\">%s</property>", tmp1, tmp2);
			g_free(tmp1);
			g_free(tmp2);

			g_value_unset(&tmp_value);
		}
	}
}

static void dump_properties(GObject* obj, GParamSpec** spec, guint nproperties)
{
	guint i;
	const char* prop_name;
	char* value_content;
	char* tmp1;
	
	for (i = 0; i < nproperties; ++i)
	{
		GValue value = { 0, };
		prop_name = g_param_spec_get_name(spec[i]);
		
		g_value_init(&value, spec[i]->value_type);
		g_object_get_property(obj, prop_name, &value);

		if (!g_param_value_defaults(spec[i], &value) && g_value_type_transformable(G_VALUE_TYPE(&value), G_TYPE_STRING))
		{
			GValue tmp_value = { 0, };

			g_value_init(&tmp_value, G_TYPE_STRING);
			g_value_transform(&value, &tmp_value);

			value_content = xml_quote(g_value_get_string(&tmp_value));
			tmp1 = convert_property(prop_name);

			if (!tmp1 || !value_content)
				/* Not enough memory, the user is in problems */
				print("<!-- NOT ENOUGH MEMORY.  YOU'RE IN PROBLEMS! -->");
			else
				print("<property name=\"%s\">%s</property>", tmp1, value_content);
			
			g_free(tmp1);
			g_free(value_content);
			
			g_value_unset(&tmp_value);
		}
	}
}

static void dump(GtkWidget* widget, gpointer unused)
{
	const char* widget_name;
	char* object_name;
	guint nproperties;
	GParamSpec** spec;
	GObjectClass* oclass;
	GtkWidget* parent;
	
	g_return_if_fail(widget);

	register_internal_children(widget);
	
	oclass = (GObjectClass*) GTK_WIDGET_GET_CLASS(widget);
	/* TODO: loop on all the properties */
	spec = g_object_class_list_properties(oclass, &nproperties);

	widget_name = g_type_name(G_TYPE_FROM_INSTANCE(widget));
	object_name = get_object_name(widget_name);

	if (!object_name)
	{
		g_warning("Not enough memory");
		return;
	}

	/* dump the widget class and name */
	print("<widget class=\"%s\" id=\"%s\">", widget_name, object_name);
	++indent;
	
	g_free(object_name);

	dump_properties(G_OBJECT(widget), spec, nproperties);
	g_free(spec);
	
	/* if the widget is a container and it has children, dump its children */
	if (GTK_IS_CONTAINER(widget) && gtk_container_get_children(GTK_CONTAINER(widget)))
	{
		GList* children = g_list_copy(gtk_container_get_children(GTK_CONTAINER(widget)));
		GList* tmp = children;
		
		if (GTK_IS_BUTTON_BOX(widget))
			children = g_list_reverse(children);

		while (children != NULL)
		{
			const char* internal_name;
			
			print("");

			internal_name = g_hash_table_lookup(internal_children, GTK_WIDGET(children->data));
			if (internal_name)
				print("<child internal-child=\"%s\">", internal_name);
			else
				print("<child>");
			
			++indent;
			dump(GTK_WIDGET(children->data), NULL);
			--indent;

			print("</child>");

			children = g_list_next(children);
		}

		g_list_free(tmp);
	}

	/* end the widget dump */
	--indent;
	print("</widget>");

	/* dump the packing properties */
	parent = gtk_widget_get_parent(widget);
	if (parent)
	{
		GObjectClass* parent_class = (GObjectClass*) GTK_WIDGET_GET_CLASS(parent);
		spec = gtk_container_class_list_child_properties(parent_class, &nproperties);

		if (nproperties > 0)
		{
			print("<packing>");
			++indent;
			dump_packing_properties(widget, spec, nproperties);
			--indent;
			print("</packing>\n");
		}

		g_free(spec);
	}
}

static void dump_gdk_window(GdkWindow* window, gpointer unused)
{
	gpointer data;
	gdk_window_get_user_data(window, &data);
	dump(GTK_WIDGET(data), NULL);
}

#ifndef TEST_STANDALONE
SIntercept *pptInterceptFuncs[] = 
   { 0x0 };

int _init()
{
	return 0;
}
#endif

int edalg_detect(const char* filename)
{
	GList* toplevels = gdk_window_get_toplevels();
	widget_count = g_hash_table_new_full(g_str_hash, g_str_equal, (GDestroyNotify) g_free, NULL);
	internal_children = g_hash_table_new_full(g_str_hash, g_str_equal, NULL, NULL);
	
	output_file = fopen(filename, "w");
	if (!output_file)
	{
		printf("Failed to open \"%s\" at [%p].\n", filename, filename);
		output_file = stdout;
	}

	print("<?xml version=\"1.0\" standalone=\"no\"?> <!--*- mode: xml -*-->");
	print("<!DOCTYPE glade-interface SYSTEM \"http://glade.gnome.org/glade-2.0.dtd\">");
	print("\n<glade-interface>\n");

	g_list_foreach(toplevels, (GFunc) dump_gdk_window, NULL);

	print("\n</glade-interface>");

	if (output_file != stdout)
		fclose(output_file);
	
	g_hash_table_destroy(widget_count);
	g_hash_table_destroy(internal_children);
	
	return 0;
}
